use super::{Height, Width};
use std::os::unix::io::RawFd;
use rustix::fd::BorrowedFd;

/// Returns the size of the terminal defaulting to STDOUT, if available.
///
/// If STDOUT is not a tty, returns `None`
pub fn terminal_size() -> Option<(Width, Height)> {
    terminal_size_using_fd(rustix::stdio::raw_stdout())
}

/// Returns the size of the terminal using the given file descriptor, if available.
///
/// If the given file descriptor is not a tty, returns `None`
pub fn terminal_size_using_fd(fd: RawFd) -> Option<(Width, Height)> {
    use rustix::termios::{isatty, tcgetwinsize};

    // TODO: Once I/O safety is stabilized, the enlosing function here should
    // be unsafe due to taking a `RawFd`. We should then move the main
    // logic here into a new function which takes a `BorrowedFd` and is safe.
    let fd = unsafe { BorrowedFd::borrow_raw(fd) };

    if !isatty(fd) {
        return None;
    }

    let winsize = tcgetwinsize(fd).ok()?;

    let rows = winsize.ws_row;
    let cols = winsize.ws_col;

    if rows > 0 && cols > 0 {
        Some((Width(cols), Height(rows)))
    } else {
        None
    }
}

#[test]
/// Compare with the output of `stty size`
fn compare_with_stty() {
    use std::process::Command;
    use std::process::Stdio;

    //in debian the tests run without a terminal, create one for testing
    let slavename;
    let slavefd;
    use libc::{posix_openpt,O_RDWR,O_NOCTTY,grantpt,unlockpt,ptsname_r,winsize,ioctl,TIOCSWINSZ,open,c_char};
    let mut slavenamearr:[c_char;1024] = [0;1024];
    unsafe {
        use std::ffi::CStr;
        let masterfd = posix_openpt(O_RDWR | O_NOCTTY);
        if masterfd < 0 {panic!("posix_openpt failed");}
        if grantpt(masterfd) < 0 {panic!("grantpt failed");}
        if unlockpt(masterfd) < 0 {panic!("unlockpt failed");}
        if ptsname_r(masterfd,slavenamearr.as_mut_ptr(),1024) != 0 {panic!("ptsname_r failed");}
        slavename = CStr::from_ptr(slavenamearr.as_mut_ptr()).to_str().unwrap();
        let ws = winsize {
            ws_row: 123,
            ws_col: 456,
            ws_xpixel: 0,
            ws_ypixel: 0,
        };
        if ioctl(masterfd,TIOCSWINSZ,&ws) != 0 {panic!("TIOCSWINSZ failed")}
        slavefd = open(slavenamearr.as_mut_ptr(), O_RDWR|O_NOCTTY);
    }

    let (rows, cols) = if cfg!(target_os = "illumos") {
        // illumos stty(1) does not accept a device argument, instead using
        // stdin unconditionally:
        let output = Command::new("stty")
            .stdin(Stdio::inherit())
            .output()
            .unwrap();
        assert!(output.status.success());

        // stdout includes the row and columns thus: "rows = 80; columns = 24;"
        let vals = String::from_utf8(output.stdout)
            .unwrap()
            .lines()
            .map(|line| {
                // Split each line on semicolons to get "k = v" strings:
                line.split(';')
                    .map(str::trim)
                    .map(str::to_string)
                    .collect::<Vec<_>>()
            })
            .flatten()
            .filter_map(|term| {
                // split each "k = v" string and look for rows/columns:
                match term.splitn(2, " = ").collect::<Vec<_>>().as_slice() {
                    ["rows", n] | ["columns", n] => Some(n.parse().unwrap()),
                    _ => None,
                }
            })
            .collect::<Vec<_>>();
        (vals[0], vals[1])
    } else {
        let output = if cfg!(target_os = "linux") {
            Command::new("stty")
                .arg("size")
                .arg("-F")
                .arg(slavename)
                .stderr(Stdio::inherit())
                .output()
                .unwrap()
        } else {
            Command::new("stty")
                .arg("-f")
                .arg(slavename)
                .arg("size")
                .stderr(Stdio::inherit())
                .output()
                .unwrap()
        };

        assert!(output.status.success());
        let stdout = String::from_utf8(output.stdout).unwrap();
        // stdout is "rows cols"
        let mut data = stdout.split_whitespace();
        println!("{}", stdout);
        let rows = u16::from_str_radix(data.next().unwrap(), 10).unwrap();
        let cols = u16::from_str_radix(data.next().unwrap(), 10).unwrap();
        (rows, cols)
    };
    println!("{} {}", rows, cols);

    if let Some((Width(w), Height(h))) = terminal_size_using_fd(slavefd) {
        assert_eq!(rows, h);
        assert_eq!(cols, w);
    } else {
        panic!("terminal_size() return None");
    }
}
